#!/usr/bin/env python3
from unittest.mock import patch

from django.contrib.gis.geos import LinearRing
from django.core.management.base import BaseCommand
from django.utils import timezone as DjangoTimeZone

from arkindex.documents.models import Corpus, Element, MetaData, MetaType
from arkindex.images.models import Image, ImageServer
from arkindex.ponos.models import Farm
from arkindex.process.models import (
    FeatureUsage,
    Process,
    ProcessMode,
    Repository,
    Worker,
    WorkerType,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tools import fake_now
from arkindex.training.models import DatasetSet
from arkindex.users.models import Group, Right, Role, User


def square_polygon(x1, x2):
    return LinearRing((x1, x1), (x2, x1), (x2, x2), (x1, x2), (x1, x1))


class Command(BaseCommand):
    help = "Create test objects in the database"

    @patch.object(DjangoTimeZone, "now", fake_now)
    def handle(self, *args, **options):

        # Create 1 image server
        imgsrv = ImageServer.objects.create(
            display_name="Test Server",
            url="http://server",
            s3_bucket="iiif",
        )

        # Create 6 images
        img1 = Image.objects.create(path="img1", width=1000, height=1000, server=imgsrv)
        img2 = Image.objects.create(path="img2", width=1000, height=1000, server=imgsrv)
        img3 = Image.objects.create(path="img3", width=1000, height=1000, server=imgsrv)
        img4 = Image.objects.create(path="img4", width=1000, height=1000, server=imgsrv)
        img5 = Image.objects.create(path="img5", width=1000, height=1000, server=imgsrv)
        img6 = Image.objects.create(path="img6", width=1000, height=1000, server=imgsrv)

        # Create an admin and a normal user
        superuser = User.objects.create_superuser("root@root.fr", "Pa$$w0rd", display_name="Admin")
        superuser.verified_email = True
        superuser.save()

        user = User.objects.create_user("user@user.fr", "Pa$$w0rd", display_name="Test user")
        user.verified_email = True
        user.save()

        # Create a group owned by the user with 2 other members
        group = Group.objects.create(name="User group", public=False)
        group.add_member(user=user, level=Role.Admin.value)
        group.add_member(
            user=User.objects.create(email="user2@user.fr", display_name="Test user write", verified_email=True),
            level=Role.Contributor.value
        )
        group.add_member(
            user=User.objects.create(email="user3@user.fr", display_name="Test user read", verified_email=True),
            level=Role.Guest.value
        )

        # Create a GitLab worker repository
        gitlab_repo = Repository.objects.create(
            url="http://gitlab/repo",
        )

        # Create a revision on this repository
        revision = gitlab_repo.revisions.create(
            hash="42",
            message="Salve",
            author="Some user"
        )

        # Create another worker repository
        worker_repo = Repository.objects.create(
            url="http://my_repo.fake/workers/worker",
        )

        # Create a revision on this repository
        revision = worker_repo.revisions.create(
            hash="1337",
            message="My w0rk3r",
            author="Test user"
        )

        # Create worker types
        dla_worker_type = WorkerType.objects.create(slug="dla", display_name="Document Layout Analysis")
        recognizer_worker_type = WorkerType.objects.create(slug="recognizer", display_name="Recognizer")
        gpu_worker_type = WorkerType.objects.create(slug="worker", display_name="Worker requiring a GPU")
        import_worker_type = WorkerType.objects.create(slug="import", display_name="Import")
        custom_worker_type = WorkerType.objects.create(slug="custom", display_name="Custom")
        init_type = WorkerType.objects.create(slug="init", display_name="Elements Initialisation")

        farm = Farm.objects.create(name="Wheat farm")
        farm.memberships.create(user=user, level=Role.Guest.value)

        # Create the elements initialisation worker version
        init_worker = WorkerVersion.objects.create(
            worker=Worker.objects.create(
                name="Elements Initialisation Worker",
                slug="initialisation",
                type=init_type,
            ),
            revision=None,
            version=1,
            configuration={
                "docker": {
                    "command": "worker-init-elements"
                }
            },
            state=WorkerVersionState.Available,
            docker_image_iid="registry.gitlab.teklia.com/arkindex/workers/init-elements:latest"
        )

        # Create some workers with available versions
        recognizer_worker = WorkerVersion.objects.create(
            worker=worker_repo.workers.create(
                name="Recognizer",
                slug="reco",
                type=recognizer_worker_type,
            ),
            revision=revision,
            configuration={"test": 42},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.somewhere.com/something:latest"
        )
        dla_worker = WorkerVersion.objects.create(
            worker=worker_repo.workers.create(
                name="Document layout analyser",
                slug="dla",
                type=dla_worker_type,
            ),
            revision=revision,
            configuration={"test": 42},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.somewhere.com/something:latest"
        )

        WorkerVersion.objects.create(
            worker=worker_repo.workers.create(
                name="File import",
                slug="file_import",
                type=import_worker_type,
            ),
            revision=revision,
            configuration={},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.somewhere.com/something:latest"
        )

        WorkerVersion.objects.create(
            worker=worker_repo.workers.create(
                name="Worker requiring a GPU",
                slug="worker-gpu",
                type=gpu_worker_type,
            ),
            revision=revision,
            configuration={"test": 42},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.somewhere.com/something:latest",
            gpu_usage=FeatureUsage.Required
        )

        # Create a generic worker and its version that uses a ML Model
        WorkerVersion.objects.create(
            worker=worker_repo.workers.create(
                name="Generic worker with a Model",
                slug="generic",
                type=recognizer_worker_type,
            ),
            revision=revision,
            configuration={"test": 42},
            state=WorkerVersionState.Available,
            gpu_usage=FeatureUsage.Disabled,
            model_usage=FeatureUsage.Required,
            docker_image_iid="registry.somewhere.com/something:latest"
        )

        # Create a custom worker version that is not linked to a Git repository/revision
        custom_version = WorkerVersion.objects.create(
            worker=Worker.objects.create(
                name="Custom worker",
                slug="custom",
                type=custom_worker_type,
            ),
            revision=None,
            version=1,
            configuration={"custom": "value"},
            state=WorkerVersionState.Created,
        )

        # Create a local process for both the user and superuser with a WorkerRun
        user_local_process = Process.objects.create(
            mode=ProcessMode.Local,
            creator=user,
        )
        superuser_local_process = Process.objects.create(
            mode=ProcessMode.Local,
            creator=superuser,
        )
        user_local_process.worker_runs.create(
            version=custom_version,
            parents=[],
        )
        superuser_local_process.worker_runs.create(
            version=custom_version,
            parents=[],
        )

        # Create a corpus
        corpus = Corpus.objects.create(
            name="Unit Tests",
            public=True
        )

        # Give rights to the user
        Right.objects.create(user=user, content_object=corpus, level=Role.Admin.value)

        # Create a Workers process with worker runs on this corpus
        process = Process.objects.create(
            farm=farm,
            mode=ProcessMode.Workers,
            creator=user,
            corpus=corpus,
            name="Process fixture",
        )
        init_worker_run = process.worker_runs.create(
            version=init_worker,
            parents=[]
        )
        dla_worker_run = process.worker_runs.create(
            version=dla_worker,
            parents=[init_worker_run.id],
        )
        process.worker_runs.create(
            version=recognizer_worker,
            parents=[dla_worker_run.id],
        )

        # Create allowed metadata
        corpus.allowed_metadatas.create(type=MetaType.Text, name="folio")
        corpus.allowed_metadatas.create(type=MetaType.Date, name="date")
        corpus.allowed_metadatas.create(type=MetaType.Location, name="location")

        # Create entity types
        corpus.entity_types.create(name="location")
        corpus.entity_types.create(name="person")
        corpus.entity_types.create(name="organization")
        corpus.entity_types.create(name="date")
        corpus.entity_types.create(name="number")

        # Create element types
        volume_type = corpus.types.create(
            slug="volume",
            display_name="Volume",
            folder=True,
        )
        page_type = corpus.types.create(
            slug="page",
            display_name="Page",
        )
        act_type = corpus.types.create(
            slug="act",
            display_name="Act",
        )
        surface_type = corpus.types.create(
            slug="surface",
            display_name="Surface",
        )
        line_type = corpus.types.create(
            slug="text_line",
            display_name="Line",
        )
        word_type = corpus.types.create(
            slug="word",
            display_name="Word",
        )

        # Create 2 datasets
        dataset_1 = corpus.datasets.create(name="First Dataset", description="dataset number one", creator=user)
        dataset_2 = corpus.datasets.create(name="Second Dataset", description="dataset number two", creator=user)
        # Create their sets
        DatasetSet.objects.bulk_create(
            DatasetSet(name=name, dataset_id=dataset_1.id) for name in ["train", "dev", "test"]
        )
        DatasetSet.objects.bulk_create(
            DatasetSet(name=name, dataset_id=dataset_2.id) for name in ["train", "dev", "test"]
        )

        # Create 2 volumes
        vol1 = Element.objects.create(
            corpus=corpus,
            type=volume_type,
            name="Volume 1",
        )
        vol2 = Element.objects.create(
            corpus=corpus,
            type=volume_type,
            name="Volume 2",
        )

        # Create 3 pages in each volume
        p1_1 = Element.objects.create(
            corpus=corpus,
            type=page_type,
            name="Volume 1, page 1r",
            image=img1,
            polygon=square_polygon(0, 1000),
        )
        p1_2 = Element.objects.create(
            corpus=corpus,
            type=page_type,
            name="Volume 1, page 1v",
            image=img2,
            polygon=square_polygon(0, 1000),
        )
        p1_3 = Element.objects.create(
            corpus=corpus,
            type=page_type,
            name="Volume 1, page 2r",
            image=img3,
            polygon=square_polygon(0, 1000),
        )
        p1_1.metadatas.set([MetaData(name="folio", value="1r", type=MetaType.Text)], bulk=False)
        p1_2.metadatas.set([MetaData(name="folio", value="1v", type=MetaType.Text)], bulk=False)
        p1_3.metadatas.set([MetaData(name="folio", value="2r", type=MetaType.Text)], bulk=False)
        p1_1.add_parent(vol1)
        p1_2.add_parent(vol1)
        p1_3.add_parent(vol1)

        p2_1 = Element.objects.create(
            corpus=corpus,
            type=page_type,
            name="Volume 2, page 1r",
            image=img4,
            polygon=square_polygon(0, 1000),
        )
        p2_2 = Element.objects.create(
            corpus=corpus,
            type=page_type,
            name="Volume 2, page 1v",
            image=img5,
            polygon=square_polygon(0, 1000),
        )
        p2_3 = Element.objects.create(
            corpus=corpus,
            type=page_type,
            name="Volume 2, page 2r",
            image=img6,
            polygon=square_polygon(0, 1000),
        )
        p2_1.metadatas.set([MetaData(name="folio", value="1r", type=MetaType.Text)], bulk=False)
        p2_2.metadatas.set([MetaData(name="folio", value="1v", type=MetaType.Text)], bulk=False)
        p2_3.metadatas.set([MetaData(name="folio", value="2r", type=MetaType.Text)], bulk=False)
        p2_1.add_parent(vol2)
        p2_2.add_parent(vol2)
        p2_3.add_parent(vol2)

        # Create transcriptions on images of volume 1
        for page in (p1_1, p1_2, p1_3):
            for word, pos in [("PARIS", 100), ("ROY", 400), ("DATUM", 700)]:
                element = corpus.elements.create(
                    type=word_type,
                    name=word,
                    image=page.image,
                    polygon=square_polygon(pos, pos + 100)
                )
                element.add_parent(page)
                element.transcriptions.create(
                    worker_version=recognizer_worker,
                    text=word,
                    confidence=1.0,
                )

        # Create a page transcription on page 1
        p1_1.transcriptions.create(
            worker_version=recognizer_worker,
            text="Lorem ipsum dolor sit amet",
            confidence=1.0,
        )

        # Create a text line on the first page of volume 1
        text_line = Element.objects.create(
            corpus=corpus,
            type=line_type,
            name="Text line",
            image=img1,
            polygon=square_polygon(400, 500),
        )
        text_line.add_parent(p1_1)

        # Create 5 acts on volume 1
        act1 = Element.objects.create(corpus=corpus, type=act_type, name="Act 1")
        act2 = Element.objects.create(corpus=corpus, type=act_type, name="Act 2")
        act3 = Element.objects.create(corpus=corpus, type=act_type, name="Act 3")
        act4 = Element.objects.create(corpus=corpus, type=act_type, name="Act 4")
        act5 = Element.objects.create(corpus=corpus, type=act_type, name="Act 5")
        act1.metadatas.set([MetaData(name="number", value="1", type=MetaType.Text)], bulk=False)
        act2.metadatas.set([MetaData(name="number", value="2", type=MetaType.Text)], bulk=False)
        act3.metadatas.set([MetaData(name="number", value="3", type=MetaType.Text)], bulk=False)
        act4.metadatas.set([MetaData(name="number", value="4", type=MetaType.Text)], bulk=False)
        act5.metadatas.set([MetaData(name="number", value="5", type=MetaType.Text)], bulk=False)
        act1.add_parent(vol1)
        act2.add_parent(vol1)
        act3.add_parent(vol1)
        act4.add_parent(vol1)
        act5.add_parent(vol1)

        # Create surfaces on the 3 pages of volume 1
        # Page 1 [A B] Page 2 [C] Page 3 [D E F]
        sa = Element.objects.create(
            corpus=corpus,
            type=surface_type,
            name="Surface A",
            image=img1,
            polygon=square_polygon(0, 600),
        )
        sb = Element.objects.create(
            corpus=corpus,
            type=surface_type,
            name="Surface B",
            image=img1,
            polygon=square_polygon(600, 1000),
        )
        sc = Element.objects.create(
            corpus=corpus,
            type=surface_type,
            name="Surface C",
            image=img2,
            polygon=square_polygon(0, 1000),
        )
        sd = Element.objects.create(
            corpus=corpus,
            type=surface_type,
            name="Surface D",
            image=img3,
            polygon=square_polygon(0, 300),
        )
        se = Element.objects.create(
            corpus=corpus,
            type=surface_type,
            name="Surface E",
            image=img3,
            polygon=square_polygon(300, 600),
        )
        sf = Element.objects.create(
            corpus=corpus,
            type=surface_type,
            name="Surface F",
            image=img3,
            polygon=square_polygon(600, 1000),
        )
        # Act 1 [A] Act 2 [B C] Act 3 [D] Act 4 [E] Act 5 [F]
        sa.add_parent(act1)
        sb.add_parent(act2)
        sc.add_parent(act2)
        sd.add_parent(act3)
        se.add_parent(act4)
        sf.add_parent(act5)

        # Create paths for top-level elements
        for element in Element.objects.filter(paths__isnull=True):
            element.paths.create(path=[])
