import uuid
from datetime import datetime, timedelta, timezone

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Element, EntityType, MetaType, Transcription
from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User


class TestChildrenElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.worker_version = WorkerVersion.objects.get(worker__slug="dla")
        cls.worker_run = WorkerRun.objects.get(version=cls.worker_version, process__mode=ProcessMode.Workers)
        cls.page = cls.corpus.elements.get(name="Volume 1, page 1r")

    def test_element_children(self):
        lonely_element = Element.objects.get(name="Volume 2, page 2r")
        response = self.client.get(reverse("api:elements-children", kwargs={"pk": str(lonely_element.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()["results"], [])

        response = self.client.get(reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()["results"]), 8)
        self.assertSetEqual(
            {r["type"] for r in response.json()["results"]},
            {"act", "page"},
        )

        response = self.client.get(
            reverse("api:elements-children", kwargs={"pk": str(self.vol.id)})
            + "?recursive&type=surface"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            [r["name"] for r in response.json()["results"]],
            [
                "Surface A",
                "Surface B",
                "Surface C",
                "Surface D",
                "Surface E",
                "Surface F"
            ],
        )

        response = self.client.get(reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}) + "?recursive")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()["results"]), 20)
        self.assertSetEqual(
            {r["type"] for r in response.json()["results"]},
            {"act", "page", "surface", "text_line", "word"},
        )

        response = self.client.get(reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}) + "?type=page")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r["name"] for r in response.json()["results"]],
            [
                "Volume 1, page 1r",
                "Volume 1, page 1v",
                "Volume 1, page 2r"
            ]
        )

        response = self.client.get(reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}) + "?name=2")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r["name"] for r in response.json()["results"]],
            [
                "Volume 1, page 2r",
                "Act 2",
            ]
        )

        nested_volume = self.corpus.elements.create(
            type=self.vol.type,
            name="Nested volume",
        )
        nested_volume.add_parent(self.vol)

        response = self.client.get(reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}) + "?folder")
        self.assertListEqual(response.json()["results"], [
            {
                "id": str(nested_volume.id),
                "name": nested_volume.name,
                "type": nested_volume.type.slug,
                "worker_version_id": None,
                "worker_run": None,
                "confidence": None,
                "zone": None,
                "rotation_angle": 0,
                "mirrored": False,
                "created": nested_volume.created.isoformat().replace("+00:00", "Z"),
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "classes": None,
                "has_children": None,
                "metadata": None,
                "thumbnail_url": nested_volume.thumbnail.s3_url,
            }
        ])

        nested_volume.worker_version = self.worker_version
        nested_volume.save()
        response = self.client.get(reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}) + "?folder")
        self.assertListEqual(response.json()["results"], [
            {
                "id": str(nested_volume.id),
                "name": nested_volume.name,
                "type": nested_volume.type.slug,
                "worker_version_id": str(self.worker_version.id),
                "worker_run": None,
                "confidence": None,
                "zone": None,
                "rotation_angle": 0,
                "mirrored": False,
                "created": nested_volume.created.isoformat().replace("+00:00", "Z"),
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "classes": None,
                "has_children": None,
                "metadata": None,
                "thumbnail_url": nested_volume.thumbnail.s3_url,
            }
        ])

    def test_element_children_with_zone_true(self):
        lonely_element = Element.objects.get(name="Volume 1, page 2r")
        response = self.client.get(
            reverse("api:elements-children", kwargs={"pk": str(self.vol.id)})
            + "?with_zone=true&type=page&name=2"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()["results"], [
            {
                "id": str(lonely_element.id),
                "name": lonely_element.name,
                "type": lonely_element.type.slug,
                "worker_version_id": None,
                "worker_run": None,
                "confidence": None,
                "zone": {
                    "id": str(lonely_element.id),
                    "image" : {
                        "id": str(lonely_element.image.id),
                        "height": lonely_element.image.height,
                        "width": lonely_element.image.width,
                        "path": lonely_element.image.path,
                        "s3_url": None,
                        "server": {
                            "display_name": lonely_element.image.server.display_name,
                            "max_height": None,
                            "max_width": None,
                            "url": lonely_element.image.server.url,
                        },
                        "status": lonely_element.image.status.value,
                        "url": lonely_element.image.url
                    },
                    "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
                    "url": lonely_element.iiif_url,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "created": lonely_element.created.isoformat().replace("+00:00", "Z"),
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "classes": None,
                "has_children": None,
                "metadata": None,
                "thumbnail_url": None,
            }
        ])

    def test_element_children_with_zone_false(self):
        lonely_element = Element.objects.get(name="Volume 1, page 2r")
        response = self.client.get(
            reverse("api:elements-children", kwargs={"pk": str(self.vol.id)})
            + "?with_zone=false&type=page&name=2"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()["results"], [
            {
                "id": str(lonely_element.id),
                "name": lonely_element.name,
                "type": lonely_element.type.slug,
                "worker_version_id": None,
                "worker_run": None,
                "confidence": None,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "created": lonely_element.created.isoformat().replace("+00:00", "Z"),
                "classes": None,
                "has_children": None,
                "metadata": None,
                "thumbnail_url": lonely_element.iiif_thumbnail_url,
            }
        ])

    def test_element_children_with_corpus_false(self):
        lonely_element = Element.objects.get(name="Volume 1, page 2r")
        response = self.client.get(
            reverse("api:elements-children", kwargs={"pk": str(self.vol.id)})
            + "?with_corpus=false&type=page&name=2"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()["results"], [
            {
                "id": str(lonely_element.id),
                "name": lonely_element.name,
                "type": lonely_element.type.slug,
                "worker_version_id": None,
                "worker_run": None,
                "confidence": None,
                "zone": {
                    "id": str(lonely_element.id),
                    "image" : {
                        "id": str(lonely_element.image.id),
                        "height": lonely_element.image.height,
                        "width": lonely_element.image.width,
                        "path": lonely_element.image.path,
                        "s3_url": None,
                        "server": {
                            "display_name": lonely_element.image.server.display_name,
                            "max_height": None,
                            "max_width": None,
                            "url": lonely_element.image.server.url,
                        },
                        "status": lonely_element.image.status.value,
                        "url": lonely_element.image.url
                    },
                    "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
                    "url": lonely_element.iiif_url,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "created": lonely_element.created.isoformat().replace("+00:00", "Z"),
                "classes": None,
                "has_children": None,
                "metadata": None,
                "thumbnail_url": None,
            }
        ])

    def test_element_children_with_worker_run(self):
        new_element = Element.objects.create(name="gru", type=self.page.type, corpus=self.corpus)
        worker_run_child = Element.objects.create(name="bob", type=self.page.type, worker_run=self.worker_run, worker_version=self.worker_version, corpus=self.corpus)
        worker_run_child.add_parent(new_element)
        response = self.client.get(
            reverse("api:elements-children", kwargs={"pk": str(new_element.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()["results"], [
            {
                "id": str(worker_run_child.id),
                "name": "bob",
                "type": "page",
                "worker_version_id": str(self.worker_version.id),
                "worker_run": {
                    "id": str(self.worker_run.id),
                    "summary": self.worker_run.summary
                },
                "zone": None,
                "confidence": None,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "created": worker_run_child.created.isoformat().replace("+00:00", "Z"),
                "classes": None,
                "has_children": None,
                "metadata": None,
                "thumbnail_url": worker_run_child.iiif_thumbnail_url,
            }
        ])

    def test_element_children_worker_run_filter(self):
        worker_run_child = Element.objects.create(name="bob", type=self.page.type, worker_run=self.worker_run, worker_version=self.worker_version, corpus=self.corpus)
        worker_run_child.add_parent(self.vol)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"worker_run": str(self.worker_run.id)}
            )

        self.assertListEqual(
            [r["name"] for r in response.json()["results"]],
            [
                "bob",
            ]
        )

    def test_element_children_worker_version_filter(self):
        self.corpus.elements.filter(name__contains="page 1r").update(worker_version=self.worker_version)

        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"worker_version": str(self.worker_version.id)}
            )

        self.assertListEqual(
            [r["name"] for r in response.json()["results"]],
            [
                "Volume 1, page 1r",
            ]
        )

    def test_element_children_worker_run_worker_version_exclusive(self):
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"worker_run": str(self.worker_run.id), "worker_version": str(self.worker_version.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"__all__": ["The worker run and worker version filters are mutually exclusive."]})

    def test_element_children_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"worker_version": "blah"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_version": ["Invalid UUID"]})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"worker_version": uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_version": ["This worker version does not exist."]})

    def test_element_children_worker_run_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"worker_run": "blah"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_run": ["Invalid UUID"]})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"worker_run": uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"worker_run": ["This worker run does not exist."]})

    def test_element_children_ordering(self):
        expected = [
            "Act 4",
            "Act 5",
            "Act 1",
            "Act 2",
            "Act 3",
        ]
        offset = Element.objects.filter(paths__path__last=self.vol.id).exclude(type__slug="act").count()
        for ordering, name in enumerate(expected):
            Element.objects.get(type__slug="act", name=name).paths.update(ordering=ordering + offset)

        response = self.client.get(
            reverse("api:elements-children", kwargs={"pk": str(self.vol.id)})
            + "?recursive&type=act"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r["name"] for r in response.json()["results"]],
            expected,
        )

    def test_children_with_has_children(self):
        # Create a childless element
        element = self.corpus.elements.create(
            name="Spearow",
            type=self.corpus.types.get(slug="surface")
        )
        element.add_parent(self.vol)

        with self.assertNumQueries(7):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"with_has_children": True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element["name"]: element["has_children"] for element in response.json()["results"]},
            {
                "Act 1": True,
                "Act 2": True,
                "Act 3": True,
                "Act 4": True,
                "Act 5": True,
                "Volume 1, page 1r": True,
                "Volume 1, page 1v": True,
                "Volume 1, page 2r": True,
                "Spearow": False,
            }
        )

    def test_element_children_no_duplicating(self):
        """
        Create a weird element structure to check that the multiple paths do not result
        in duplicated elements when listing element children:
                Page
                  |
                 / \
                /   \
           Table_1  Table_2           col1 col2
            |  |    /                 ┌───┬───┐
            |  |   /             row1 │ A │ B │
          Row Column                  ├───┼───┤
            |  |                 row2 │ C │ D │
            Cell                      └───┴───┘
        """
        self.client.force_login(self.user)
        self.corpus.types.create(slug="table", display_name="Table")
        self.corpus.types.create(slug="column", display_name="Column")
        self.corpus.types.create(slug="row", display_name="Row")
        self.corpus.types.create(slug="cell", display_name="Cell")

        page = Element.objects.create(corpus=self.corpus, type=self.page.type, image=self.page.image, name="The Test Page", polygon=self.page.polygon)
        page.image.status = S3FileStatus.Checked
        page.image.save()

        def _create(type_slug, name, parent, x, y, width, height):
            response = self.client.post(reverse("api:elements-create"), data={
                "corpus": str(self.corpus.id),
                "type": type_slug,
                "name": name,
                "parent": str(parent.id),
                "image": str(page.image.id),
                "polygon": [
                    [x, y],
                    [x, y + height],
                    [x + width, y + height],
                    [x + width, y],
                    [x, y]
                ],
            }, format="json")
            self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
            return self.corpus.elements.get(id=response.json()["id"])

        def _add_parent(parent, child):
            response = self.client.post(reverse("api:element-parent", kwargs={
                "parent": str(parent.id),
                "child": str(child.id),
            }))
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        table_1 = _create("table", "The Table", page, 100, 100, 200, 200)
        table_2 = _create("table", "The Other Table", page, 200, 100, 200, 200)

        row1 = _create("row", "Row 1", table_1, 100, 100, 200, 100)
        row2 = _create("row", "Row 2", table_1, 100, 200, 200, 100)

        col1 = _create("column", "Column 1", table_1, 100, 100, 100, 200)
        col2 = _create("column", "Column 2", table_1, 100, 100, 100, 200)

        cell_a = _create("cell", "Cell A", row1, 100, 100, 100, 100)
        cell_b = _create("cell", "Cell B", row1, 200, 100, 100, 100)
        cell_c = _create("cell", "Cell C", row2, 100, 200, 100, 100)
        cell_d = _create("cell", "Cell D", row2, 200, 200, 100, 100)

        _add_parent(table_2, col1)
        _add_parent(col1, cell_a)
        _add_parent(col2, cell_b)
        _add_parent(col1, cell_c)
        _add_parent(col2, cell_d)

        with self.assertNumQueries(7):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(page.id)}),
                {"recursive": True},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        # col1, cell_b and cell_c are duplicated due to a bug in Element.objects.get_descending
        # See https://gitlab.teklia.com/arkindex/backend/-/issues/769
        self.assertCountEqual([item["id"] for item in response.json()["results"]], [
            str(table_1.id),
            str(table_2.id),
            str(row1.id),
            str(row2.id),
            str(col1.id),
            str(col1.id),
            str(col2.id),
            str(cell_a.id),
            str(cell_b.id),
            str(cell_b.id),
            str(cell_c.id),
            str(cell_c.id),
            str(cell_d.id)
        ])

    def test_children_modified_since_bad_format(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                HTTP_IF_MODIFIED_SINCE="AAAAAAAAAAAAAAAAA",
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"If-Modified-Since": ["Bad date format"]})

    def test_children_modified_since_not_modified(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                HTTP_IF_MODIFIED_SINCE="Thu, 02 Apr 2099 13:37:42 GMT",
            )
        self.assertEqual(response.status_code, status.HTTP_304_NOT_MODIFIED)

    def test_children_modified_since(self):
        self.corpus.elements.filter(name="Volume 1, page 1r").update(updated="2099-04-02T13:37:43Z")
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                HTTP_IF_MODIFIED_SINCE="Thu, 02 Apr 2099 13:37:42 GMT",
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 8)
        self.assertSetEqual(
            {r["type"] for r in data["results"]},
            {"act", "page"},
        )

    def test_children_filter_metadata_name(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"metadata_name": "folio"}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element["name"] for element in response.json()["results"]],
            [
                "Volume 1, page 1r",
                "Volume 1, page 1v",
                "Volume 1, page 2r",
            ]
        )

    def test_children_filter_metadata_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"metadata_value": "1v"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "metadata_value": ["This filter is not supported without metadata_name."],
        })

    def test_children_filter_metadata_name_value(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"metadata_name": "folio", "metadata_value": "1v"}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element["name"] for element in response.json()["results"]],
            ["Volume 1, page 1v"]
        )

    def test_children_filter_metadata_operator(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"metadata_operator": "gt"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "metadata_operator": ["This option is not supported without metadata_value."]
        })

    def test_children_filter_metadata_operator_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"metadata_operator": "gt", "metadata_value": 4}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "metadata_value": ["This filter is not supported without metadata_name."],
        })

    def test_children_filter_metadata_operator_invalid(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"metadata_operator": "aaaaaa"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "metadata_operator": [
                "This metadata operator is not supported.",
                "This option is not supported without metadata_value.",
            ]
        })

    def test_children_filter_numeric_metadata_invalid_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={"metadata_name": "a", "metadata_value": "lol", "metadata_operator": "gt"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "metadata_value": ["Value must be a float when using a numeric comparison operator."],
        })

    def test_children_filter_numeric_metadata(self):
        element_type = self.corpus.types.first()
        for i in range(1, 6):
            element = self.corpus.elements.create(type=element_type, name=f"Element {i}")
            element.add_parent(self.vol)
            element.metadatas.create(name="number", type=MetaType.Numeric, value=str(i))
            element.metadatas.create(name="number", type=MetaType.Text, value="a")

        # operator, value, expected element names
        parameters = [
            ("lt", 2, ["Element 1"]),
            ("lte", 2, ["Element 1", "Element 2"]),
            ("gt", 3, ["Element 4", "Element 5"]),
            ("gte", 3, ["Element 3", "Element 4", "Element 5"]),
        ]

        for operator, value, expected_elements in parameters:
            with self.subTest(operator=operator, value=value):
                with self.assertNumQueries(4):
                    response = self.client.get(
                        reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                        data={
                            "metadata_name": "number",
                            "metadata_value": value,
                            "metadata_operator": operator,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element["name"] for element in response.json()["results"]],
                    expected_elements
                )

    def test_children_filter_metadata_contains(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={
                    "metadata_name": "folio",
                    "metadata_operator": "contains",
                    "metadata_value": "1"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element["name"] for element in response.json()["results"]],
            [
                "Volume 1, page 1r",
                "Volume 1, page 1v",
            ]
        )

    def test_children_filter_creator_email(self):
        element_type = self.corpus.types.first()
        user2 = User.objects.get(email="user2@user.fr")
        test_element = self.corpus.elements.create(type=element_type, name="Test Element", creator=user2)
        test_element.add_parent(self.vol)

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={
                    "creator_email": "user2@user.fr"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()["results"]), 1)
        self.assertEqual(response.json()["results"][0]["name"], "Test Element")

    def test_children_filter_creator_email_case_insensitive(self):
        element_type = self.corpus.types.first()
        user2 = User.objects.get(email="user2@user.fr")
        test_element = self.corpus.elements.create(type=element_type, name="Test Element", creator=user2)
        test_element.add_parent(self.vol)

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={
                    "creator_email": "USEr2@user.fr"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()["results"]), 1)
        self.assertEqual(response.json()["results"][0]["name"], "Test Element")

    def test_children_filter_creator_email_does_not_exist(self):
        """
        If there is no user with the given email, there is no error but no results are returned
        """
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                data={
                    "creator_email": "ohno@user.fr"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()["results"]), 0)

    def test_children_filter_rotation_angle(self):
        element_type = self.corpus.types.first()
        not_rotated = self.corpus.elements.create(type=element_type, name="not rotated")
        rotated = self.corpus.elements.create(type=element_type, name="rotated", rotation_angle=180)
        not_rotated.add_parent(self.vol)
        rotated.add_parent(self.vol)

        cases = [
            (0, ["not rotated"]),
            (1, []),
            (180, ["rotated"]),
            (181, [])
        ]
        for rotation_angle, expected_elements in cases:
            with self.subTest(rotation_angle=rotation_angle):
                # Requests that return no elements only make 2 SQL queries
                num_queries = 4 if len(expected_elements) else 2
                with self.assertNumQueries(num_queries):
                    response = self.client.get(
                        reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                        data={
                            "name": "rotated",
                            "rotation_angle": rotation_angle,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element["name"] for element in response.json()["results"]],
                    expected_elements
                )

    def test_children_filter_rotation_angle_invalid(self):
        cases = [
            ("to the left", "invalid literal for int() with base 10: 'to the left'"),
            (-1, "A rotation angle must be between 0 and 359 degrees"),
            (360, "A rotation angle must be between 0 and 359 degrees"),
            (54.2, "invalid literal for int() with base 10: '54.2'"),
            (4.14j, "invalid literal for int() with base 10: '4.14j'"),
        ]
        for rotation_angle, expected_error in cases:
            with self.subTest(rotation_angle=rotation_angle):
                with self.assertNumQueries(1):
                    response = self.client.get(
                        reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                        data={
                            "rotation_angle": rotation_angle,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {
                    "rotation_angle": [expected_error]
                })

    def test_children_filter_mirrored(self):
        element_type = self.corpus.types.first()
        mirrored = self.corpus.elements.create(type=element_type, name="mirrored", mirrored=True)
        not_mirrored = self.corpus.elements.create(type=element_type, name="not mirrored", mirrored=False)
        mirrored.add_parent(self.vol)
        not_mirrored.add_parent(self.vol)

        cases = [
            (False, ["not mirrored"]),
            (True, ["mirrored"]),
        ]
        for mirrored, expected_elements in cases:
            with self.subTest(mirrored=mirrored):
                with self.assertNumQueries(4):
                    response = self.client.get(
                        reverse("api:elements-children", kwargs={"pk": str(self.vol.id)}),
                        data={
                            "name": "mirrored",
                            "mirrored": mirrored,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element["name"] for element in response.json()["results"]],
                    expected_elements
                )

    def test_children_invalid_order(self):
        cases = [
            ({"order": "blah", "order_direction": "asc"}, {"order": ["Unknown sorting field"]}),
            ({"order": "name", "order_direction": "left"}, {"order_direction": ["Unknown sorting direction"]}),
            ({"order": "blah", "order_direction": "left"}, {"order": ["Unknown sorting field"]}),
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    params,
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected)

    def test_children_order_created(self):
        elements = list(Element.objects.get_descending(self.vol.id).order_by("?").only("id"))
        created = datetime(2020, 2, 2, tzinfo=timezone.utc)
        for element in elements:
            element.created = created
            created += timedelta(seconds=1)
        Element.objects.bulk_update(elements, ["created"])

        cases = [
            ("asc", elements[:20]),
            ("desc", reversed(elements[-20:])),
        ]
        for direction, expected in cases:
            with self.subTest(direction=direction), self.assertNumQueries(5):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    {"order": "created", "order_direction": direction, "recursive": True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element["id"] for element in response.json()["results"]],
                    [str(element.id) for element in expected]
                )

    def test_children_order_name(self):
        names = list(Element.objects.get_descending(self.vol.id).order_by("name", "id").values_list("name", flat=True))

        # Ensure the default position ordering returns a different ordering than by name;
        # otherwise, this test would be useless
        self.assertNotEqual(
            list(Element.objects.get_descending(self.vol.id).values_list("name", flat=True)),
            names,
        )

        cases = [
            ("asc", names[:20]),
            ("desc", list(reversed(names[-20:]))),
        ]
        for direction, expected in cases:
            with self.subTest(direction=direction), self.assertNumQueries(5):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    {"order": "name", "order_direction": direction, "recursive": True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element["name"] for element in response.json()["results"]],
                    expected
                )

    def test_children_order_position(self):
        ids = list(map(
            str,
            Element
            .objects
            .get_descending(self.vol.id)
            .order_by("paths__ordering", "id")
            .values_list("id", flat=True)
        ))

        cases = [
            ("asc", ids[:20]),
            ("desc", list(reversed(ids[-20:]))),
        ]
        for direction, expected in cases:
            with self.subTest(direction=direction), self.assertNumQueries(5):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    {"order": "position", "order_direction": direction, "recursive": True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element["id"] for element in response.json()["results"]],
                    expected
                )

    def test_children_order_random(self):
        # Get all expected element IDs as strings
        element_ids = list(map(
            str,
            Element.objects
                   .get_descending(self.vol.id)
                   .filter(paths__path__last=self.vol.id)
                   .values_list("id", flat=True)
        ))

        # Repeat the API request multiple times; the element order should change at least once
        responses = set()
        for _ in range(5):
            with self.assertNumQueries(5):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    {"order": "random"},
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)

            data = response.json()

            # Checking for the expected element IDs will start failing randomly
            # if there is more than 1 page of results.
            self.assertIsNone(data["next"])
            actual_ids = [element["id"] for element in data["results"]]
            self.assertCountEqual(element_ids, actual_ids)

            responses.add(tuple(actual_ids))

        # We know that all responses have the expected elements, so if the responses set
        # has more than one item, the ordering has changed between requests, so it is likely to be random.
        self.assertGreater(len(responses), 1)

    def test_children_with_metadata(self):
        # Remove all metadata from a page; we should get an empty list
        page1r = self.corpus.elements.get(name="Volume 1, page 1r")
        page1r.metadatas.all().delete()

        # Add an entity to a metadata; it should be excluded
        page1v = self.corpus.elements.get(name="Volume 1, page 1v")
        entity_meta = page1v.metadatas.get()
        location_type = EntityType.objects.get(name="location", corpus=self.corpus)
        entity_meta.entity = self.corpus.entities.create(name="Something", type=location_type)
        entity_meta.save()

        page2r = self.corpus.elements.get(name="Volume 1, page 2r")
        page2r_meta = page2r.metadatas.get()

        # Add another metadata, because we still need to have at least two non-empty lists
        # in this test to ensure only one query is made to prefetch all metadata of all elements
        page1v_meta = page1v.metadatas.create(name="Another metadata", type=MetaType.Text, value="something")

        with self.assertNumQueries(7):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                {"with_metadata": True, "type": "page"},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual({
            element["name"]: element["metadata"]
            for element in response.json()["results"]
        }, {
            "Volume 1, page 1r": [],
            "Volume 1, page 1v": [
                {
                    "id": str(page1v_meta.id),
                    "name": "Another metadata",
                    "type": "text",
                    "value": "something",
                    "dates": []
                }
            ],
            "Volume 1, page 2r": [
                {
                    "id": str(page2r_meta.id),
                    "name": "folio",
                    "type": "text",
                    "value": "2r",
                    "dates": [],
                }
            ],
        })

    def test_transcription_confidence_invalid(self):
        cases = [
            ({"transcription_confidence": "very"}, {"transcription_confidence": ["could not convert string to float: 'very'"]}),
            ({"transcription_confidence": "nan"}, {"transcription_confidence": ["Confidence must be between 0 and 1"]}),
            ({"transcription_confidence": "inf"}, {"transcription_confidence": ["Confidence must be between 0 and 1"]}),
            ({"transcription_confidence": "42"}, {"transcription_confidence": ["Confidence must be between 0 and 1"]}),
            ({"transcription_confidence": "-.5"}, {"transcription_confidence": ["Confidence must be between 0 and 1"]}),
            (
                {"transcription_confidence_operator": "hah"},
                {"transcription_confidence_operator": [
                    "This operator is not supported.",
                    "This option is not supported without transcription_confidence.",
                ]}
            ),
            (
                {"transcription_confidence": "0.3", "transcription_confidence_operator": "lol"},
                {"transcription_confidence_operator": ["This operator is not supported."]}
            )
        ]
        for data, expected_errors in cases:
            with self.subTest(**data):
                response = self.client.get(reverse("api:elements-children", kwargs={"pk": self.vol.id}), data)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_transcription_confidence_filter(self):
        # All transcriptions in the fixtures have a confidence of 1. We update them to be able to test properly.
        self.assertEqual(Transcription.objects.all().count(), 10)
        for i, transcription in enumerate(Transcription.objects.order_by("text")):
            transcription.confidence = i / 10
            transcription.save()

        cases = [
            (
                {"transcription_confidence": 0.5},
                ["PARIS"]
            ),
            (
                {"transcription_confidence": 0.51},
                []
            ),
            (
                {"transcription_confidence": 0.5, "transcription_confidence_operator": "gt"},
                ["PARIS", "ROY", "ROY", "ROY"]
            ),
            (
                {"transcription_confidence": 0.5, "transcription_confidence_operator": "lte"},
                ["Volume 1, page 1r", "PARIS", "PARIS", "DATUM", "DATUM", "DATUM"]
            )
        ]
        for data, element_names in cases:
            with self.subTest(**data):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    {"recursive": True, **data}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertCountEqual(
                    [element["name"] for element in response.json()["results"]],
                    element_names,
                )

    def test_transcription_worker_version_invalid(self):
        cases = [
            ("a", {"transcription_worker_version": ["Invalid UUID"]}),
            (42, {"transcription_worker_version": ["Invalid UUID"]}),
            (True, {"transcription_worker_version": ["Invalid UUID"]}),
            ("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", {"transcription_worker_version": ["This worker version does not exist."]}),
        ]
        for transcription_worker_version, expected_errors in cases:
            with self.subTest(transcription_worker_version=transcription_worker_version):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    {"transcription_worker_version": transcription_worker_version}
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_transcription_worker_version_filter(self):
        recognizer = WorkerVersion.objects.get(worker__slug="reco")
        Transcription.objects.filter(text="PARIS").update(worker_version=None)
        Transcription.objects.filter(text="ROY").update(worker_version=self.worker_version)

        cases = [
            (str(recognizer.id), ["Volume 1, page 1r", "DATUM", "DATUM", "DATUM"]),
            (str(self.worker_version.id), ["ROY", "ROY", "ROY"]),
            (False, ["PARIS", "PARIS", "PARIS"]),
        ]
        for transcription_worker_version, element_names in cases:
            with self.subTest(transcription_worker_version=transcription_worker_version):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    {"transcription_worker_version": transcription_worker_version, "recursive": True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element["name"] for element in response.json()["results"]],
                    element_names,
                )

    def test_transcription_worker_run_filter(self):
        Transcription.objects.filter(text="PARIS").update(worker_run=None)
        Transcription.objects.filter(text="ROY").update(worker_run=self.worker_run)

        cases = [
            (str(self.worker_run.id), ["ROY", "ROY", "ROY"]),
            (False, ["Volume 1, page 1r", "PARIS", "PARIS", "PARIS", "DATUM", "DATUM", "DATUM"]),
        ]
        for transcription_worker_run, element_names in cases:
            with self.subTest(transcription_worker_run=transcription_worker_run):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    {"transcription_worker_run": transcription_worker_run, "recursive": True}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertCountEqual(
                    [element["name"] for element in response.json()["results"]],
                    element_names,
                )

    def test_transcription_worker_run_invalid(self):
        cases = [
            ("a", {"transcription_worker_run": ["Invalid UUID"]}),
            (42, {"transcription_worker_run": ["Invalid UUID"]}),
            (True, {"transcription_worker_run": ["Invalid UUID"]}),
            ("aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", {"transcription_worker_run": ["This worker run does not exist."]}),
        ]
        for transcription_worker_run, expected_errors in cases:
            with self.subTest(transcription_worker_run=transcription_worker_run):
                response = self.client.get(
                    reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                    {"transcription_worker_run": transcription_worker_run}
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_transcription_worker_run_worker_version_exclusive(self):
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:elements-children", kwargs={"pk": self.vol.id}),
                {"transcription_worker_run": str(self.worker_run.id), "transcription_worker_version": str(self.worker_version.id), "recursive": True}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"__all__": ["The transcription worker run and transcription worker version filters are mutually exclusive."]})

    def test_confidence_invalid(self):
        cases = [
            ({"confidence": "very"}, {"confidence": ["could not convert string to float: 'very'"]}),
            ({"confidence": "nan"}, {"confidence": ["Confidence must be between 0 and 1"]}),
            ({"confidence": "inf"}, {"confidence": ["Confidence must be between 0 and 1"]}),
            ({"confidence": "42"}, {"confidence": ["Confidence must be between 0 and 1"]}),
            ({"confidence": "-.5"}, {"confidence": ["Confidence must be between 0 and 1"]}),
            (
                {"confidence_operator": "hah"},
                {"confidence_operator": [
                    "This operator is not supported.",
                    "This option is not supported without `confidence`.",
                ]}
            ),
            (
                {"confidence": "0.3", "confidence_operator": "lol"},
                {"confidence_operator": ["This operator is not supported."]}
            )
        ]
        for data, expected_errors in cases:
            with self.subTest(**data):
                response = self.client.get(reverse("api:elements-children", kwargs={"pk": self.vol.id}), data)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_confidence_filter(self):
        elements = Element.objects.bulk_create([
            Element(
                corpus=self.corpus,
                type=self.corpus.types.get(slug="volume"),
                confidence=i / 10,
                name=f"Confidence {i / 10}"
            )
            for i in range(11)
        ])
        for element in elements:
            element.add_parent(self.vol)

        cases = [
            (
                {"confidence": 0.5},
                ["Confidence 0.5"]
            ),
            (
                {"confidence": 0.51},
                []
            ),
            (
                {"confidence": 0.5, "confidence_operator": "gt"},
                ["Confidence 0.6", "Confidence 0.7", "Confidence 0.8", "Confidence 0.9", "Confidence 1.0"]
            ),
            (
                {"confidence": 0.5, "confidence_operator": "lte"},
                ["Confidence 0.0", "Confidence 0.1", "Confidence 0.2", "Confidence 0.3", "Confidence 0.4", "Confidence 0.5"]
            )
        ]
        for data, element_names in cases:
            with self.subTest(**data):
                response = self.client.get(reverse("api:elements-children", kwargs={"pk": self.vol.id}), data)
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element["name"] for element in response.json()["results"]],
                    element_names,
                )
