import uuid
from unittest.mock import call, patch

from django.contrib.auth.models import AnonymousUser
from django.contrib.gis.geos import LinearRing
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Entity, EntityType, MetaType, TranscriptionEntity
from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestEntitiesAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name="private")
        cls.worker_version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_version_2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.custom_version = WorkerVersion.objects.get(worker__slug="custom")
        cls.local_worker_run = WorkerRun.objects.get(version=cls.custom_version, process__creator=cls.user)
        cls.worker_run_1 = cls.worker_version_1.worker_runs.filter(process__mode=ProcessMode.Workers).get()
        cls.worker_run_2 = cls.worker_version_2.worker_runs.filter(process__mode=ProcessMode.Workers).get()
        cls.page = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.element_type = cls.corpus.types.get(slug="text_line")
        cls.person_type = EntityType.objects.get(name="person", corpus=cls.corpus)
        cls.location_type = EntityType.objects.get(name="location", corpus=cls.corpus)
        cls.org_type = EntityType.objects.get(name="organization", corpus=cls.corpus)

    def setUp(self):
        super().setUp()

        self.entity = Entity.objects.create(
            type=self.person_type,
            corpus=self.corpus,
            name="entity 1",
            worker_version=self.worker_version_1,
        )
        self.entity_bis = Entity.objects.create(
            type=self.location_type,
            corpus=self.corpus,
            name="entity 2",
            worker_version=self.worker_version_2,
        )
        self.element = self.corpus.elements.create(
            type=self.element_type,
            name="Transcription",
            image=self.page.image,
            polygon=LinearRing((0, 0), (0, 42), (42, 42), (42, 0), (0, 0)),
            worker_version=self.worker_version_1,
        )
        self.transcription = self.element.transcriptions.create(
            text="Some transcribed text",
            worker_version=self.worker_version_1,
        )
        self.metadata = self.element.metadatas.create(
            name="test 1",
            type=MetaType.Text,
            value="Blah",
            element=self.element,
            entity=self.entity,
        )
        self.transcriptionentity = TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity_bis,
            offset=2,
            length=len(self.entity_bis.name),
            worker_version=self.worker_version_1
        )
        self.tr_entities_sample = {
            "entity": str(self.entity.id),
            "offset": 4,
            "length": len(self.entity.name)
        }
        self.tr_entities_confidence_sample = {
            "entity": str(self.entity.id),
            "offset": 4,
            "length": len(self.entity.name),
            "confidence": 1.0
        }

    def test_get_entity(self):
        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:entity-details", kwargs={"pk": str(self.entity.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["id"], str(self.entity.id))
        self.assertEqual(data["type"]["id"], str(self.entity.type.id))
        self.assertEqual(data["corpus"]["id"], str(self.corpus.id))
        self.assertEqual(data["name"], self.entity.name)

    def test_get_entity_elements(self):
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:entity-elements", kwargs={"pk": str(self.entity.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        results = data["results"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]["id"], str(self.element.id))
        self.assertEqual(results[0]["name"], self.element.name)

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_get_entity_elements_corpus_acl(self, filter_rights_mock):
        # One query runs because Django >=4.2 no longer ignores unions that are known to be empty
        # https://code.djangoproject.com/ticket/35559
        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:entity-elements", kwargs={"pk": str(self.entity.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json().get("results"), [])
        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(AnonymousUser(), Corpus, Role.Guest.value),
            call(AnonymousUser(), Corpus, Role.Guest.value),
        ])

    def test_get_entity_elements_from_transcription(self):
        elt = self.corpus.elements.create(
            type=self.corpus.types.get(slug="act"),
            name="001",
            image=self.imgsrv.images.first(),
            polygon=LinearRing((0, 0), (0, 200), (200, 200), (200, 0), (0, 0)),
        )
        elt_tr = elt.transcriptions.create(worker_version=self.worker_version_1, text="goodbye")
        TranscriptionEntity.objects.create(transcription=elt_tr, entity=self.entity, offset=42, length=7)

        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:entity-elements", kwargs={"pk": str(self.entity.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        results = data["results"]
        self.assertEqual(len(results), 2)
        self.assertListEqual(
            results,
            [{
                "id": str(e.id),
                "type": e.type.slug,
                "name": e.name,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public
                },
                "thumbnail_url": None,
                "zone": {
                    "id": str(e.id),
                    "url": e.iiif_url,
                    "polygon": [
                        list(map(int, point))
                        for point in e.polygon.coords
                    ],
                    "image": {
                        "id": str(e.image.id),
                        "width": e.image.width,
                        "height": e.image.height,
                        "path": e.image.path,
                        "url": e.image.url,
                        "s3_url": None,
                        "status": e.image.status.value,
                        "server": {
                            "display_name": "Test Server",
                            "max_height": None,
                            "max_width": None,
                            "url": "http://server",
                        }
                    }
                },
                "rotation_angle": 0,
                "mirrored": False,
            } for e in (elt, self.element)]
        )

    def test_create_entity_type_invalid(self):
        data = {
            "name": "entity",
            "type_id": "location",
            "corpus": str(self.corpus.id),
            "metas": {
                "key": "value",
                "other key": "other value"
            },
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:entity-create"), data=data, format="json")
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"type_id": ["“location” is not a valid UUID."]})

    def test_create_entity_type_does_not_exist(self):
        data = {
            "name": "entity",
            "type_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "corpus": str(self.corpus.id),
            "metas": {
                "key": "value",
                "other key": "other value"
            },
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:entity-create"), data=data, format="json")
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"type_id": ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.']})

    def test_create_entity_type_does_not_exist_in_corpus(self):
        extraneous_type = EntityType.objects.create(name="Author", corpus=self.private_corpus)
        data = {
            "name": "entity",
            "type_id": str(extraneous_type.id),
            "corpus": str(self.corpus.id),
            "metas": {
                "key": "value",
                "other key": "other value"
            },
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:entity-create"), data=data, format="json")
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"type_id": [f"EntityType {str(extraneous_type.id)} does not exist in corpus Unit Tests."]})

    def test_create_entity_require_type(self):
        data = {
            "name": "entity",
            "corpus": str(self.corpus.id),
            "metas": {
                "key": "value",
                "other key": "other value"
            },
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:entity-create"), data=data, format="json")
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"type_id": ["This field is required."]})

    def test_create_entity_requires_login(self):
        data = {
            "name": "Paris",
            "type_id": str(self.location_type.id),
            "corpus": str(self.corpus.id),
            "metas": {
                "key": "value",
                "other key": "other value"
            },
        }
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:entity-create"), data=data, format="json")
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_entity_no_worker_version(self):
        """
        Only a worker run can be used to create an entity, not a worker version
        """
        data = {
            "name": "entity",
            "type_id": str(self.person_type.id),
            "corpus": str(self.corpus.id),
            "worker_version": str(self.worker_version_1.id),
            "metas": {
                "key": "value",
                "other key": "other value"
            },
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:entity-create"), data=data, format="json")
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "worker_version": ["This field is forbidden."],
        })

    def test_create_entity_with_unknown_worker_run(self):
        data = {
            "name": "entity",
            "type_id": str(self.person_type.id),
            "corpus": str(self.corpus.id),
            "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "metas": {
                "key": "value",
                "other key": "other value"
            },
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(reverse("api:entity-create"), data=data, format="json")
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"worker_run_id": ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.']})

    def test_create_entity(self):
        """
        A user can create an entity without a worker run
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:entity-create"),
                format="json",
                data={
                    "name": "test",
                    "type_id": str(self.person_type.id),
                    "corpus": str(self.corpus.id),
                    "metas": {"pokemon": "Wo-Chien"},
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        entity = Entity.objects.get(id=data["id"])
        self.assertEqual(entity.name, "test")
        self.assertEqual(entity.type, self.person_type)
        self.assertEqual(entity.corpus, self.corpus)
        self.assertFalse(entity.validated)
        self.assertEqual(entity.metas, {"pokemon": "Wo-Chien"})
        self.assertEqual(entity.worker_version_id, None)
        self.assertEqual(entity.worker_run, None)

    def test_create_entity_worker_run_non_local(self):
        """
        A regular user cannot create a transcription entity with a WorkerRun of a non-local process
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:entity-create"),
                format="json",
                data={
                    "name": "test",
                    "type_id": str(self.person_type.id),
                    "corpus": str(self.corpus.id),
                    "worker_run_id": str(self.worker_run_1.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Ponos task authentication is required to use a WorkerRun "
                "of a process other than the user's local process."
            ]

        })

    def test_create_entity_worker_run_other_user(self):
        """
        A regular user cannot create an entity with a WorkerRun of someone else's local process
        """
        private_worker_run = self.user.processes.get(mode=ProcessMode.Local).worker_runs.first()
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:entity-create"),
                format="json",
                data={
                    "name": "test",
                    "type_id": str(self.person_type.id),
                    "corpus": str(self.corpus.id),
                    "worker_run_id": str(private_worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Ponos task authentication is required to use a WorkerRun "
                "of a process other than the user's local process."
            ]

        })

    def test_create_entity_worker_run_other_process(self):
        """
        A Ponos task cannot create an entity with a WorkerRun of another process
        """
        process2 = self.worker_run_1.process.creator.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
        )
        other_worker_run = process2.worker_runs.create(version=self.worker_version_1, parents=[])
        self.worker_run_1.process.run()
        task = self.worker_run_1.process.tasks.first()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:entity-create"),
                format="json",
                data={
                    "name": "Bunnelby",
                    "type_id": str(self.person_type.id),
                    "corpus": str(self.corpus.id),
                    "worker_run_id": str(other_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Only the WorkerRuns of the authenticated task's process may be used."
            ]
        })

    def test_create_entity_local_worker_run(self):
        """
        A regular user can create an entity with a WorkerRun of their own local process
        """
        data = {
            "name": "entity",
            "type_id": str(self.person_type.id),
            "corpus": str(self.corpus.id),
            "worker_run_id": str(self.local_worker_run.id),
            "metas": {
                "key": "value",
                "other key": "other value"
            },
        }
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:entity-create"), data=data, format="json")
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        entity = Entity.objects.get(id=data["id"])
        self.assertEqual(entity.name, "entity")
        self.assertEqual(entity.type, self.person_type)
        self.assertEqual(entity.corpus, self.corpus)
        self.assertFalse(entity.validated)
        self.assertEqual(entity.metas, {
            "key": "value",
            "other key": "other value",
        })
        self.assertEqual(entity.worker_version_id, self.local_worker_run.version_id)
        self.assertEqual(entity.worker_run, self.local_worker_run)

        self.assertDictEqual(data, {
            "id": str(entity.id),
            "name": "entity",
            "type": {
                "id": str(self.person_type.id),
                "name": "person",
                "color": "ff0000",
            },
            "metas": {
                "key": "value",
                "other key": "other value",
            },
            "corpus": {
                "id": str(self.corpus.id),
                "name": "Unit Tests",
                "public": True,
            },
            "validated": False,
            "worker_run": {
                "id": str(self.local_worker_run.id),
                "summary": self.local_worker_run.summary,
            },
            "worker_version_id": str(self.local_worker_run.version_id),
        })

    def test_create_entity_task_auth(self):
        """
        An entity can be created with a WorkerRun of a non-local process
        when authenticated as a Ponos task of this process
        """
        self.worker_run_1.process.run()
        task = self.worker_run_1.process.tasks.first()

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:entity-create"),
                format="json",
                data={
                    "name": "test",
                    "type_id": str(self.person_type.id),
                    "corpus": str(self.corpus.id),
                    "worker_run_id": str(self.worker_run_1.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        entity = Entity.objects.get(id=data["id"])
        self.assertEqual(entity.name, "test")
        self.assertEqual(entity.type, self.person_type)
        self.assertEqual(entity.corpus, self.corpus)
        self.assertFalse(entity.validated)
        self.assertEqual(entity.metas, None)
        self.assertEqual(entity.worker_version, self.worker_version_1)
        self.assertEqual(entity.worker_run, self.worker_run_1)

    def test_create_entity_worker_run_local_task_auth(self):
        """
        An entity can be created with a WorkerRun of a Local process
        even when authenticated as a Ponos task from a different process
        """
        local_process = self.user.processes.get(mode=ProcessMode.Local)
        local_worker_run = local_process.worker_runs.get()

        self.worker_run_1.process.run()
        task = self.worker_run_1.process.tasks.first()

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:entity-create"),
                format="json",
                data={
                    "name": "Bunnelby",
                    "type_id": str(self.person_type.id),
                    "corpus": str(self.corpus.id),
                    "worker_run_id": str(local_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        entity = Entity.objects.get(id=data["id"])
        self.assertEqual(entity.name, "Bunnelby")
        self.assertEqual(entity.type, self.person_type)
        self.assertEqual(entity.corpus, self.corpus)
        self.assertFalse(entity.validated)
        self.assertEqual(entity.metas, None)
        self.assertEqual(entity.worker_version_id, local_worker_run.version_id)
        self.assertEqual(entity.worker_run, local_worker_run)

    def test_create_transcription_entity(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data=self.tr_entities_sample,
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        transcription_entity = TranscriptionEntity.objects.get(
            transcription=self.transcription,
            entity=self.entity,
            offset=self.tr_entities_sample["offset"],
            length=self.tr_entities_sample["length"]
        )
        self.assertIsNotNone(transcription_entity)
        self.assertDictEqual(
            response.json(),
            {
                "entity": str(transcription_entity.entity.id),
                "offset": transcription_entity.offset,
                "length": transcription_entity.length,
                "worker_run": None,
                "confidence": None
            }
        )

    def test_create_transcription_entity_with_confidence(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data=self.tr_entities_confidence_sample,
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        transcription_entity = TranscriptionEntity.objects.get(
            transcription=self.transcription,
            entity=self.entity,
            offset=self.tr_entities_confidence_sample["offset"],
            length=self.tr_entities_confidence_sample["length"],
            confidence=self.tr_entities_confidence_sample["confidence"]
        )
        self.assertIsNotNone(transcription_entity)
        self.assertDictEqual(
            response.json(),
            {
                "entity": str(transcription_entity.entity.id),
                "offset": transcription_entity.offset,
                "length": transcription_entity.length,
                "worker_run": None,
                "confidence": transcription_entity.confidence
            }
        )

    def test_create_transcription_entity_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data=self.tr_entities_sample,
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_transcription_entity_worker_version(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": len(self.entity.name),
                    "worker_version_id": str(self.worker_version_1.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_version_id": ["This field is forbidden."],
        })

    def test_create_transcription_entity_worker_run(self):
        """
        A regular user can create classifications with a WorkerRun of their own local process
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                    "worker_run_id": str(self.local_worker_run.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {
            "entity": str(self.entity.id),
            "offset": 4,
            "length": 8,
            "worker_run": {
                "id": str(self.local_worker_run.id),
                "summary": self.local_worker_run.summary
            },
            "confidence": None,
        })

    def test_create_transcription_entity_forbidden_version(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                    "worker_version_id": str(self.worker_version_1.id),
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_version_id": ["This field is forbidden."],
        })

    def test_create_transcription_entity_bad_worker_run(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                    "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.'],
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_create_transcription_entity_wrong_acl(self, filter_rights_mock):
        self.client.force_login(self.user)
        self.element.corpus = self.private_corpus
        self.element.save()
        filter_rights_mock.return_value = Corpus.objects.exclude(id=self.private_corpus.id)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data=self.tr_entities_sample,
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(
            response.json(),
            {"transcription": ["invalid UUID or Corpus write-access is forbidden"]}
        )
        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Corpus, Role.Contributor.value),
            call(self.user, Corpus, Role.Contributor.value),
        ])

    def test_create_transcription_entity_wrong_length(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 0
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"length": ["Ensure this value is greater than or equal to 1."]}
        )

    def test_create_transcription_entity_wrong_high_confidence(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 1,
                    "confidence": 2.0
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"confidence": ["Ensure this value is less than or equal to 1."]}
        )

    def test_create_transcription_entity_wrong_low_confidence(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 1,
                    "confidence": -0.5
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"confidence": ["Ensure this value is greater than or equal to 0."]}
        )

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_create_transcription_entity_different_corpus(self, filter_rights_mock):
        self.client.force_login(self.user)
        person_type = EntityType.objects.create(name="person", corpus=self.private_corpus)
        ent = Entity.objects.create(
            type=person_type,
            corpus=self.private_corpus,
            name="a private entity",
            worker_version=self.worker_version_1,
        )
        self.tr_entities_sample.update({"entity": ent.id})
        filter_rights_mock.return_value = Corpus.objects.exclude(id=self.private_corpus.id)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data=self.tr_entities_sample,
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(
            response.json(),
            {"entity": [f'Invalid pk "{ent.id}" - object does not exist.']}
        )
        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Corpus, Role.Contributor.value),
            call(self.user, Corpus, Role.Contributor.value),
        ])

    def test_create_transcription_entity_duplicate(self):
        self.client.force_login(self.user)
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data=self.tr_entities_sample,
                format="json"
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "__all__": ["This entity is already linked to this transcription at this position."]
        })

    def test_create_transcription_entity_duplicate_worker_run(self):
        """
        No duplicate entity can be created on a trancription
        """
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=8,
            worker_run=self.local_worker_run,
            worker_version=self.worker_version_1
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                    "worker_run_id": str(self.local_worker_run.id),
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "__all__": ["This entity is already linked to this transcription by this worker run at this position."]
        })

    def test_create_transcription_entity_worker_run_non_local(self):
        """
        A regular user cannot create an entity on a transcription with a WorkerRun of a non-local process
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                format="json",
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                    "worker_run_id": str(self.worker_run_1.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Ponos task authentication is required to use a WorkerRun "
                "of a process other than the user's local process."
            ]
        })

    def test_create_transcription_entity_worker_run_other_user(self):
        """
        A regular user cannot create an entity on a transcription with a WorkerRun of someone else's local process
        """
        private_worker_run = self.user.processes.get(mode=ProcessMode.Local).worker_runs.first()
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                format="json",
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                    "worker_run_id": str(private_worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Ponos task authentication is required to use a WorkerRun "
                "of a process other than the user's local process."
            ]
        })

    def test_create_transcription_entity_worker_run_other_process(self):
        """
        A Ponos task cannot create an entity on a transcription with a WorkerRun of another process
        """
        process2 = self.worker_run_1.process.creator.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
        )
        other_worker_run = process2.worker_runs.create(version=self.worker_version_1, parents=[])
        self.worker_run_1.process.run()
        task = self.worker_run_1.process.tasks.first()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                format="json",
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                    "worker_run_id": str(other_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Only the WorkerRuns of the authenticated task's process may be used."
            ]
        })

    def test_create_transcription_entity_worker_run_task_auth(self):
        """
        An entity can be created on a transcription with a WorkerRun of a non-local process
        when authenticated as a Ponos task of this process
        """
        self.worker_run_1.process.run()
        task = self.worker_run_1.process.tasks.first()

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                format="json",
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                    "worker_run_id": str(self.worker_run_1.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {
            "entity": str(self.entity.id),
            "offset": 4,
            "length": 8,
            "worker_run": {
                "id": str(self.worker_run_1.id),
                "summary": self.worker_run_1.summary
            },
            "confidence": None,
        })

    def test_create_transcription_entity_worker_run_local_task_auth(self):
        """
        An entity can be created on a transcription with a WorkerRun of a Local process
        even when authenticated as a Ponos task from a different process
        """
        local_process = self.user.processes.get(mode=ProcessMode.Local)
        local_worker_run = local_process.worker_runs.get()

        self.worker_run_1.process.run()
        task = self.worker_run_1.process.tasks.first()

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                format="json",
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                    "worker_run_id": str(local_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertDictEqual(response.json(), {
            "entity": str(self.entity.id),
            "offset": 4,
            "length": 8,
            "worker_run": {
                "id": str(local_worker_run.id),
                "summary": local_worker_run.summary
            },
            "confidence": None,
        })

    def test_create_transcription_entity_manual_existing_worker_run(self):
        """
        A manual TranscriptionEntity can be created even when one exists with the same attributes from a WorkerRun
        """
        self.client.force_login(self.user)
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=8,
            worker_run=self.worker_run_1,
            worker_version=self.worker_version_1
        )
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entity": str(self.entity.id),
                    "offset": 4,
                    "length": 8,
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        transcription_entity = self.transcription.transcription_entities.filter(worker_run=None, entity=self.entity).get()
        self.assertDictEqual(
            response.json(),
            {
                "entity": str(transcription_entity.entity.id),
                "offset": transcription_entity.offset,
                "length": transcription_entity.length,
                "worker_run": None,
                "confidence": None
            }
        )

    def test_create_transcription_entity_key_missing(self):
        self.client.force_login(self.user)
        data = self.tr_entities_sample
        del data["length"]
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entity-create", kwargs={"pk": str(self.transcription.id)}),
                data=data,
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_transcription_entities_open(self):
        response = self.client.get(reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_list_transcription_entities_wrong_acl(self, filter_rights_mock):
        self.transcription.element.corpus = self.private_corpus
        self.transcription.element.save()
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    def test_list_transcription_entities(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertListEqual(
            data["results"],
            [{
                "entity": {
                    "id": str(self.entity_bis.id),
                    "name": self.entity_bis.name,
                    "type": {
                        "name": "location",
                        "id": str(self.location_type.id),
                        "color": "ff0000"
                    },
                    "metas": None,
                    "validated": self.entity_bis.validated,
                    "worker_version_id": str(self.worker_version_2.id),
                    "worker_run": None,
                },
                "length": self.transcriptionentity.length,
                "offset": self.transcriptionentity.offset,
                "worker_version_id": str(self.worker_version_1.id),
                "worker_run": None,
                "confidence": None
            }]
        )

    def test_list_transcription_entities_superuser(self):
        self.client.force_login(self.superuser)
        self.assertFalse(self.transcription.element.corpus.memberships.filter(user=self.superuser).exists())
        response = self.client.get(reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertEqual(
            data["results"],
            [{
                "entity": {
                    "id": str(self.entity_bis.id),
                    "name": self.entity_bis.name,
                    "type": {
                        "name": "location",
                        "id": str(self.location_type.id),
                        "color": "ff0000"
                    },
                    "metas": None,
                    "validated": self.entity_bis.validated,
                    "worker_version_id": str(self.worker_version_2.id),
                    "worker_run": None,
                },
                "length": self.transcriptionentity.length,
                "offset": self.transcriptionentity.offset,
                "worker_version_id": str(self.worker_version_1.id),
                "worker_run": None,
                "confidence": None
            }],
        )

    def test_list_transcription_entities_validation(self):
        self.client.force_login(self.user)
        cases = [
            (
                {"worker_version": "blah"},
                2,
                {"worker_version": ["Invalid UUID."]},
            ),
            (
                {"worker_version": str(uuid.uuid4())},
                3,
                {"worker_version": ["This worker version does not exist."]},
            ),
            (
                {"entity_worker_version": "blah"},
                2,
                {"entity_worker_version": ["Invalid UUID."]},
            ),
            (
                {"entity_worker_version": str(uuid.uuid4())},
                3,
                {"entity_worker_version": ["This worker version does not exist."]},
            ),
            (
                {"worker_run": "blah"},
                2,
                {"worker_run": ["Invalid UUID."]},
            ),
            (
                {"worker_run": str(uuid.uuid4())},
                3,
                {"worker_run": ["This worker run does not exist."]},
            ),
            (
                {"entity_worker_run": "blah"},
                2,
                {"entity_worker_run": ["Invalid UUID."]},
            ),
            (
                {"entity_worker_run": str(uuid.uuid4())},
                3,
                {"entity_worker_run": ["This worker run does not exist."]},
            ),
            (
                {
                    "worker_version": "blah",
                    "worker_run": "blah",
                    "entity_worker_version": "blah",
                    "entity_worker_run": "blah",
                },
                2,
                {
                    "worker_version": ["Invalid UUID."],
                    "worker_run": ["Invalid UUID."],
                    "entity_worker_version": ["Invalid UUID."],
                    "entity_worker_run": ["Invalid UUID."],
                },
            ),
            (
                {
                    "worker_version": str(uuid.uuid4()),
                    "worker_run": str(uuid.uuid4()),
                    "entity_worker_version": str(uuid.uuid4()),
                    "entity_worker_run": str(uuid.uuid4()),
                },
                6,
                {
                    "worker_version": ["This worker version does not exist."],
                    "worker_run": ["This worker run does not exist."],
                    "entity_worker_version": ["This worker version does not exist."],
                    "entity_worker_run": ["This worker run does not exist."],
                },
            ),
        ]
        for parameters, num_queries, expected in cases:
            with self.subTest(parameters=parameters):
                with self.assertNumQueries(num_queries):
                    response = self.client.get(
                        reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}),
                        data=parameters,
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), expected)

    def test_list_transcription_entities_worker_version(self):
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=8,
            length=len(self.entity.name),
            worker_version=self.worker_version_2
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}),
                {"worker_version": str(self.worker_version_2.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()["results"],
            [{
                "entity": {
                    "id": str(self.entity.id),
                    "name": "entity 1",
                    "type": {
                        "name": "person",
                        "id": str(self.person_type.id),
                        "color": "ff0000"
                    },
                    "metas": None,
                    "validated": False,
                    "worker_version_id": str(self.worker_version_1.id),
                    "worker_run": None,
                },
                "length": 8,
                "offset": 8,
                "worker_version_id": str(self.worker_version_2.id),
                "worker_run": None,
                "confidence": None
            }]
        )

    def test_list_transcription_entities_worker_run(self):
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=8,
            length=len(self.entity.name),
            worker_version=self.worker_version_2,
            worker_run=self.worker_run_2,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}),
                {"worker_run": str(self.worker_run_2.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()["results"],
            [{
                "entity": {
                    "id": str(self.entity.id),
                    "name": "entity 1",
                    "type": {
                        "name": "person",
                        "id": str(self.person_type.id),
                        "color": "ff0000"
                    },
                    "metas": None,
                    "validated": False,
                    "worker_version_id": str(self.worker_version_1.id),
                    "worker_run": None,
                },
                "length": 8,
                "offset": 8,
                "worker_version_id": str(self.worker_version_2.id),
                "worker_run": {
                    "id": str(self.worker_run_2.id),
                    "summary": self.worker_run_2.summary,
                },
                "confidence": None,
            }]
        )

    def test_list_transcription_entities_manual(self):
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=2,
            length=len(self.entity.name)
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}),
                {"worker_version": False}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()["results"],
            [{
                "entity": {
                    "id": str(self.entity.id),
                    "name": "entity 1",
                    "type": {
                        "name": "person",
                        "id": str(self.person_type.id),
                        "color": "ff0000"
                    },
                    "metas": None,
                    "validated": False,
                    "worker_version_id": str(self.worker_version_1.id),
                    "worker_run": None,
                },
                "length": 8,
                "offset": 2,
                "worker_version_id": None,
                "worker_run": None,
                "confidence": None
            }]
        )

    def test_list_transcription_entities_entity_worker_version(self):
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=8,
            length=len(self.entity.name),
            worker_version=self.worker_version_2,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}),
                {"entity_worker_version": str(self.worker_version_1.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()["results"],
            [{
                "entity": {
                    "id": str(self.entity.id),
                    "name": "entity 1",
                    "type": {
                        "name": "person",
                        "id": str(self.person_type.id),
                        "color": "ff0000"
                    },
                    "metas": None,
                    "validated": False,
                    "worker_version_id": str(self.worker_version_1.id),
                    "worker_run": None,
                },
                "length": 8,
                "offset": 8,
                "worker_version_id": str(self.worker_version_2.id),
                "worker_run": None,
                "confidence": None
            }]
        )

    def test_list_transcription_entities_entity_worker_run(self):
        entity = self.corpus.entities.create(
            type=self.person_type,
            name="Some entity",
            worker_run=self.worker_run_1,
            worker_version=self.worker_version_1,
        )
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=entity,
            offset=8,
            length=len(self.entity.name),
            worker_version=self.worker_version_2,
            worker_run=self.worker_run_2,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}),
                {"entity_worker_run": str(self.worker_run_1.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()["results"],
            [{
                "entity": {
                    "id": str(entity.id),
                    "name": "Some entity",
                    "type": {
                        "name": "person",
                        "id": str(self.person_type.id),
                        "color": "ff0000"
                    },
                    "metas": None,
                    "validated": False,
                    "worker_version_id": str(self.worker_version_1.id),
                    "worker_run": {
                        "id": str(self.worker_run_1.id),
                        "summary": self.worker_run_1.summary,
                    },
                },
                "length": 8,
                "offset": 8,
                "worker_version_id": str(self.worker_version_2.id),
                "worker_run": {
                    "id": str(self.worker_run_2.id),
                    "summary": self.worker_run_2.summary,
                },
                "confidence": None
            }]
        )

    def test_list_transcription_entities_entity_manual(self):
        manual_entity = Entity.objects.create(
            type=self.person_type,
            corpus=self.corpus,
            name="manual entity"
        )
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=manual_entity,
            offset=8,
            length=len(manual_entity.name),
            worker_version=self.worker_version_2
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}),
                {"entity_worker_version": False}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()["results"],
            [{
                "entity": {
                    "id": str(manual_entity.id),
                    "name": "manual entity",
                    "type": {
                        "name": "person",
                        "id": str(self.person_type.id),
                        "color": "ff0000"
                    },
                    "metas": None,
                    "validated": False,
                    "worker_version_id": None,
                    "worker_run": None,
                },
                "length": 13,
                "offset": 8,
                "worker_version_id": str(self.worker_version_2.id),
                "worker_run": None,
                "confidence": None
            }]
        )

    def test_list_transcription_entities_worker_version_entity_worker_version(self):
        """
        Filter a transcription entity both by its own worker version, and its entity's worker version
        """
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=8,
            length=len(self.entity.name),
            worker_version=self.worker_version_2
        )
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name),
            worker_version=self.worker_version_1
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse("api:transcription-entities", kwargs={"pk": str(self.transcription.id)}),
                {"entity_worker_version": str(self.worker_version_1.id), "worker_version": str(self.worker_version_2.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()["results"],
            [{
                "entity": {
                    "id": str(self.entity.id),
                    "name": "entity 1",
                    "type": {
                        "name": "person",
                        "id": str(self.person_type.id),
                        "color": "ff0000"
                    },
                    "metas": None,
                    "validated": False,
                    "worker_version_id": str(self.worker_version_1.id),
                    "worker_run": None,
                },
                "length": 8,
                "offset": 8,
                "worker_version_id": str(self.worker_version_2.id),
                "worker_run": None,
                "confidence": None
            }]
        )

    def test_list_corpus_entities(self):
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:corpus-entities", kwargs={"pk": str(self.corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            "count": 2,
            "previous": None,
            "next": None,
            "number": 1,
            "results": [
                {
                    "id": str(self.entity.id),
                    "name": "entity 1",
                    "type": {
                        "name": "person",
                        "id": str(self.person_type.id),
                        "color": "ff0000"
                    },
                    "validated": False,
                    "metas": None,
                    "worker_version_id": str(self.worker_version_1.id),
                    "worker_run": None,
                },
                {
                    "id": str(self.entity_bis.id),
                    "name": "entity 2",
                    "type": {
                        "name": "location",
                        "id": str(self.location_type.id),
                        "color": "ff0000"
                    },
                    "validated": False,
                    "metas": None,
                    "worker_version_id": str(self.worker_version_2.id),
                    "worker_run": None,
                }
            ]
        })

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_list_corpus_entities_acl(self, has_access_mock):
        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:corpus-entities", kwargs={"pk": str(self.private_corpus.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(AnonymousUser(), self.private_corpus, Role.Guest.value, skip_public=False))

    def test_list_corpus_entities_parent(self):
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:corpus-entities", kwargs={"pk": str(self.corpus.id)}),
                {"parent": str(self.element.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 1,
            "previous": None,
            "next": None,
            "number": 1,
            "results": [
                {
                    "id": str(self.entity_bis.id),
                    "name": "entity 2",
                    "type": {
                        "name": "location",
                        "id": str(self.location_type.id),
                        "color": "ff0000"
                    },
                    "validated": False,
                    "metas": None,
                    "worker_version_id": str(self.worker_version_2.id),
                    "worker_run": None
                },
            ]
        })

    def test_list_corpus_entities_invalid_parent_uuid(self):
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:corpus-entities", kwargs={"pk": str(self.corpus.id)}), {"parent": "a"})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "parent": ["Parent ID is not a valid UUID"]
        })

    def test_list_corpus_entities_parent_not_found(self):
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:corpus-entities", kwargs={"pk": str(self.corpus.id)}),
                {"parent": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"},
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_corpus_entities_name(self):
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:corpus-entities", kwargs={"pk": str(self.corpus.id)}), {"name": "2"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            "count": 1,
            "previous": None,
            "next": None,
            "number": 1,
            "results": [
                {
                    "id": str(self.entity_bis.id),
                    "name": "entity 2",
                    "type": {
                        "name": "location",
                        "id": str(self.location_type.id),
                        "color": "ff0000"
                    },
                    "validated": False,
                    "metas": None,
                    "worker_version_id": str(self.worker_version_2.id),
                    "worker_run": None,
                }
            ]
        })

    def test_delete_entity(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse("api:entity-details", kwargs={"pk": str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Entity.DoesNotExist):
            Entity.objects.get(id=self.entity.id)

    def test_delete_entity_requires_login(self):
        response = self.client.delete(reverse("api:entity-details", kwargs={"pk": str(self.entity_bis.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_entity_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.delete(reverse("api:entity-details", kwargs={"pk": str(self.entity_bis.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    # UpdateEntity

    def test_update_entity_requires_name_and_type(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {"name": "a new name"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"type_id": ["This field is required."]})

    def test_update_validate_entity(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {
                    "name": self.entity_bis.name,
                    "type_id": str(self.location_type.id),
                    "validated": True
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.entity_bis.refresh_from_db()
        self.assertEqual(self.entity_bis.validated, True)
        self.assertEqual(self.entity_bis.moderator, self.user)

    def test_update_unvalidate_entity(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {
                    "name": self.entity_bis.name,
                    "type_id": str(self.location_type.id),
                    "validated": False
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.entity_bis.refresh_from_db()
        self.assertEqual(self.entity_bis.validated, False)
        self.assertEqual(self.entity_bis.moderator, None)

    def test_update_validate_entity_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {
                    "name": self.entity_bis.name,
                    "type_id": str(self.location_type.id),
                    "validated": True
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_unvalidate_entity_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {
                    "name": self.entity_bis.name,
                    "type_id": str(self.location_type.id),
                    "validated": False
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_validate_entity_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {
                    "name": self.entity_bis.name,
                    "type_id": str(self.location_type.id),
                    "validated": True
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_unvalidate_entity_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {
                    "name": self.entity_bis.name,
                    "type_id": str(self.location_type.id),
                    "validated": False
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_entity_change_type(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {
                    "name": self.entity_bis.name,
                    "type_id": str(self.person_type.id),
                    "validated": False
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.entity_bis.refresh_from_db()
        self.assertEqual(self.entity_bis.validated, False)
        self.assertEqual(self.entity_bis.moderator, None)
        self.assertEqual(self.entity_bis.type, self.person_type)

    # PartialUpdateEntity

    def test_partial_update_validate_entity(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {"validated": True},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.entity_bis.refresh_from_db()
        self.assertEqual(self.entity_bis.validated, True)
        self.assertEqual(self.entity_bis.moderator, self.user)

    def test_partial_update_unvalidate_entity(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {"validated": False},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.entity_bis.refresh_from_db()
        self.assertEqual(self.entity_bis.validated, False)
        self.assertEqual(self.entity_bis.moderator, None)

    def test_partial_update_validate_entity_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {"validated": True},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_unvalidate_entity_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {"validated": False},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_validate_entity_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {"validated": True},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_unvalidate_entity_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:entity-details", kwargs={"pk": self.entity_bis.id}),
                {"validated": False},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
